#include "ircd.h"
#include "channel.h"
#include "client.h"
#include "handlers.h"   
#include "hash.h"
#include "ircd.h"
#include "ircd_defs.h"
#include "ircd_string.h"
#include "ircd_alloc.h"
#include "list.h"
#include "msg.h"
#include "numnicks.h"
#include "sprintf_irc.h"
#include "s_conf.h"
#include "s_user.h"
#include "s_bsd.h"
#include "send.h"
#include "patch.h"
#include "numeric.h"
#include "s_debug.h"
#include "patchlevel.h"

#include <ctype.h>
#include <string.h>
#include <stdlib.h>


void show_patchconf(struct Client *cptr);

#ifdef OLD_ISOPTION

int IsOption(int option)
{
  struct ConfItem *aconf;
  int o;
  char ch[2];
  aconf = find_patchconf();
  if(!aconf) return 0;
  ch[0] = aconf->host[option-1];
  ch[1] = '\0';
  o = atoi(ch);
  return (o)?1:0;
}

int LogOption(int option)
{
  struct ConfItem *aconf;
  int o;
  char ch[2];
  aconf = find_patchconf();
  if(!aconf) return 0;
  ch[0] = aconf->host[option-1];
  ch[1] = '\0';
  o = atoi(ch);
  return (o>=2)?1:0;
}
#else
unsigned int eLine=0;
unsigned int eLineL1=0;
unsigned int eLineL2=0;
#endif

void LogFeature(struct Client *cptr, int type, const char *fmt, ...)
{
  struct ConfItem *aconf;
  struct Client *acptr;
  int o,i;
  char buffer[512+1];
  va_list va;
#ifdef OLD_ISOPTION
  char ch[2];
  aconf = find_patchconf();
  if(!aconf) return;
  ch[0] = aconf->host[type-1];
  ch[1] = '\0';
  o = atoi(ch);
#else
  if(eLineL1 & type) o = 2;
  if(eLineL2 & type) o = 3;
#endif
  va_start(va, fmt);
  vsnprintf(buffer, 512, fmt, va);
  switch(o)
  {
    case 2:
      sprintf_irc(sendbuf, ":%s " MSG_WALLOPS " :%s[%s@%s] %s", me.name,
                  cptr->name, cptr->user->username, cptr->user->host, buffer);
      for (i = 0; i <= HighestFd; ++i) {
      if ((acptr = LocalClientArray[i]) && !IsServer(acptr) &&
           IsAnOper(acptr) && SendWallops(acptr))
           sendbufto_one(acptr);
      }
      sendto_serv_butone(cptr, "%s " TOK_WALLOPS " :%s[%s@%s] %s", NumServ(&me), cptr->name, cptr->user->username, cptr->user->host, buffer);
      break;
    case 3:
      for(i = 0; i <= HighestFd; i++)
      {
        if((acptr = LocalClientArray[i]) && !IsServer(acptr) && SendServNotice(acptr))
          if(IsAnOper(acptr))
            sendto_one(acptr, ":%s NOTICE %s :%s[%s@%s] %s", me.name, acptr->name, cptr->name, cptr->user->username, cptr->user->host, buffer);
      }
  }
  va_end(va);
}

int m_patchinfo(struct Client *cptr, struct Client *sptr, int parc, char *parv[])
{
  struct Client *acptr;

  if (MyConnect(sptr) && parc > 1) {
    if (!(acptr = find_match_server(parv[1]))) {
      sendto_one(sptr, err_str(ERR_NOSUCHSERVER), me.name, parv[0], parv[1]);
      return 0;
    }
    parv[1] = acptr->name;
  }
  if (hunt_server(0, cptr, sptr, "%s%s " TOK_PATCHINFO " :%s", 1, parc,
      parv) != HUNTED_ISME) return 0;

  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, parv[0], HEADERLINE);
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, parv[0],
    "IRCu patch " PATCHLEVEL " by Tero Nybacka (sivis@sivis.org)");
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, parv[0],
    "This patch is derived from QuakeNET patch by Darren Coleman");
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, parv[0], HEADERLINE);
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, parv[0],
    "Additional thanks in random order: Antti Eturautti, Jere Kataja,");
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, parv[0],
    "Joe Cise, Matthew Devney, Michael Meier and rest of the QuakeNet");
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, parv[0],
    "and B2IRC operators.");
  sendto_one(sptr, rpl_str(RPL_DATASTR), me.name, parv[0], HEADERLINE);
  if(IsAnOper(sptr))
  {
    show_patchconf(sptr);
  }
  return 0;
}
 
void show_patchconf(struct Client *cptr)
{
  struct ConfItem *aconf;
  char buftext[BUFSIZE+1];
  int i=0;
  aconf = find_patchconf();
  if(!aconf) 
  {
    return;
  }
  sprintf(buftext, "Patch configuration for %s:", me.name);
  sendto_one(cptr, rpl_str(RPL_DATASTR), me.name, cptr->name, buftext);
  sprintf(buftext, "(E:%s) D:%d %d %d [*=Usage is wallopped]:", aconf->host, eLine, eLineL1, eLineL2);
  sendto_one(cptr, rpl_str(RPL_DATASTR), me.name, cptr->name, buftext);
  sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_OPER_CANSET_K, N_OPER_CANSET_K, LogOption(OPER_CANSET_K)?'*':' ', IsOption(OPER_CANSET_K)?"YES":"NO");
  sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_OPER_CANSET_X, N_OPER_CANSET_X, LogOption(OPER_CANSET_X)?'*':' ', IsOption(OPER_CANSET_X)?"YES":"NO");
  if(IsOption(OPER_CANSET_X))
  {
    sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_OPER_CANJOIN_I, N_OPER_CANJOIN_I, LogOption(OPER_CANJOIN_I)?'*':' ', IsOption(OPER_CANJOIN_I)?"YES":"NO");
    sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_OPER_CANJOIN_K, N_OPER_CANJOIN_K, LogOption(OPER_CANJOIN_K)?'*':' ', IsOption(OPER_CANJOIN_K)?"YES":"NO");
    sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_OPER_CANJOIN_B, N_OPER_CANJOIN_B, LogOption(OPER_CANJOIN_B)?'*':' ', IsOption(OPER_CANJOIN_B)?"YES":"NO");
    sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_OPER_CANJOIN_L, N_OPER_CANJOIN_L, LogOption(OPER_CANJOIN_L)?'*':' ', IsOption(OPER_CANJOIN_L)?"YES":"NO");
    sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_X_CANKICK_K, N_X_CANKICK_K, LogOption(X_CANKICK_K)?'*':' ', IsOption(X_CANKICK_K)?"YES":"NO");
    sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_X_CANDEOP_K, N_X_CANDEOP_K, LogOption(X_CANDEOP_K)?'*':' ', IsOption(X_CANDEOP_K)?"YES":"NO");
    sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_X_CANKILL_K, N_X_CANKILL_K, LogOption(X_CANKILL_K)?'*':' ', IsOption(X_CANKILL_K)?"YES":"NO");
  }
  sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_OPER_CANSEND_N, N_OPER_CANSEND_N, LogOption(OPER_CANSEND_N)?'*':' ', IsOption(OPER_CANSEND_N)?"YES":"NO");
  sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_OPER_CANSEND_M, N_OPER_CANSEND_M, LogOption(OPER_CANSEND_M)?'*':' ', IsOption(OPER_CANSEND_M)?"YES":"NO");
  sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_OPER_SERVMODE, N_OPER_SERVMODE, LogOption(OPER_SERVMODE)?'*':' ', IsOption(OPER_SERVMODE)?"YES":"NO");
  sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_OPER_DYNHOST, N_OPER_DYNHOST, LogOption(OPER_DYNHOST)?'*':' ', IsOption(OPER_DYNHOST)?"YES":"NO");
  sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_USER_SETHOST, N_USER_SETHOST, LogOption(USER_SETHOST)?'*':' ', IsOption(USER_SETHOST)?"YES":"NO");
  sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_USE_FAKEHOST, N_USE_FAKEHOST, LogOption(USE_FAKEHOST)?'*':' ', IsOption(USE_FAKEHOST)?"YES":"NO");
  sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_OPER_CANFLOOD, N_OPER_CANFLOOD, LogOption(OPER_CANFLOOD)?'*':' ', IsOption(OPER_CANFLOOD)?"YES":"NO");
  sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_OPER_NOLAG, N_OPER_NOLAG, LogOption(OPER_NOLAG)?'*':' ', IsOption(OPER_NOLAG)?"YES":"NO");
  sendto_one(cptr, rpl_str(RPL_PATCHCONF), me.name, cptr->name, DESC_OPER_NETRIDE, N_OPER_NETRIDE, LogOption(OPER_NETRIDE)?'*':' ', IsOption(OPER_NETRIDE)?"YES":"NO");
}

int IsService(struct Client *cptr)
{
  struct ConfItem *aconf;
/* Sanity checks */
  if(!IsUser(cptr)) return 0;
/* Check if the nick is juped in U:lines, if not, 
  user shouldn't be protected in any case */
  if(!isNickJuped(cptr->name)) return 0;
  for(aconf = GlobalConfList ; aconf ; aconf = aconf->next)
    if((aconf->status & CONF_UWORLD) && (strCasediff(aconf->host, cptr->user->server->name)==0) && (strCasediff(aconf->host, cptr->user->host)==0))
      return 1;
  return 0;
}

/* --------------------------------------------------------
 * **Daz** PadChar();
 *    This function takes an integer parameter (length)
 *    and returns a string of 'max' - 'length' in length
 * --------------------------------------------------------
 */
char *PadChar(int length, int max, char ch)
{
   static char str[BUFSIZE];
   unsigned int i, cnt;
   char *dst;

   i = max - length; 
   if (i < 0 || i >= BUFSIZE) return " ";

   for (dst = str, cnt = 0; cnt <= i; cnt++) *dst++ = ch;
   *dst = '\0';

   return str;  
}


/* -------------------------------------------------------
 * **Daz** strtolower()
 *    This function accepts a string (or char) and
 *    returns the same string/char in lowercase.
 * -------------------------------------------------------
 */
char *strtolower(char *str)
{
   static char buf[BUFSIZE];
   unsigned int i;
   
   if (strlen(str) > BUFSIZE)   /* Sanity check */
      return str;
   
   strcpy(buf,str);
   for (i = 0;i < (strlen(str));i++)
      buf[i] = tolower(buf[i]);
   return buf;
}


/* --------------------------------------------------------
 * **Daz** ConstructMask()
 *         Takes 'hoststr' as a parameter and constructs a
 *         standard nick!user@host mask from it, determining
 *         the desired search the user is attempting to do.
 * --------------------------------------------------------
 */
char *ConstructMask(char *hoststr)
{
   char nickm[NICKLEN + 1], userm[USERLEN + 1], hostm[HOSTLEN + 1];
   static char targhost[NICKLEN + USERLEN + HOSTLEN + 3];
   char *p = NULL;

   /* ------------------------------------------------------------------------- */
   /* We need to determine what kind of search the user wishes to do.  In order
    * to ascertain this, we have to parse the host parameter and decide
    * what parameters the user has given us.  This isn't particularly
    * easy when you consider that users very often pass parameters in totally
    * the wrong order, may miss some parameters entirely and might - either
    * maliciously or unintentionally - pass garbled data as a parameter..
    *
    * Essentially we can assume that if the user includes a '@' char, that they
    * wish to limit the search to specific usernames.  Likewise, if a '!'
    * character is supplied as well, that indicates the user wishes to further
    * restrict the search to nicknames as well.
    *
    * Don't worry if you don't understand this code, I have re-written it about
    * 10 times and I still don't like the look of it :)
    *                                                   -- Durzel.
    */
   /* ------------------------------------------------------------------------- */

   strcpy(nickm,"*");
   strcpy(userm,"*");
   strcpy(hostm,"*");

   if (!strchr(hoststr, '!') && !strchr(hoststr, '@'))
      ircd_strncpy(hostm,hoststr,HOSTLEN);
   else
   {
     if (p = strchr(hoststr, '@')) {
        *p++ = '\0';
        if (*p) ircd_strncpy(hostm,p, HOSTLEN);
     }

     /* Get the nick!user mask */
     if (p = strchr(hoststr, '!')) {
        *p++ = '\0';
        if (*p) ircd_strncpy(userm,p,USERLEN);
        if (*hoststr) ircd_strncpy(nickm,hoststr,NICKLEN);
     }
     else if (*hoststr)
        /* Durz: We should only do the following *IF* the hoststr has not already been
         * copied into hostm (ie. neither ! or @ specified).. otherwise, when we do
         * /quote check *.barrysworld.com - we end up with targhost as: *!*.barryswo@*.barrysworld.com
         */
        ircd_strncpy(userm,hoststr,USERLEN);
   }

   /* Copy formatted string into "targhost" buffer */
   sprintf_irc(targhost, "%s!%s@%s", nickm, userm, hostm);

   return targhost;
}


/* --------------------------------------------------------
 * **Daz** IsValidHost()
 *    This procedure takes as its argument a hostmask
 *    and checks to see whether or not it is valid.
 *
 *    Examples of non-valid hostmasks include any that
 *    include special ASCII characters.
 *
 *    (standard C boolean result)
 *    Returns:  -1 = TRUE  (valid host)
 *               0 = FALSE (invalid host)
 * --------------------------------------------------------
 */
signed int IsValidHost(char *host)
{
  char *s;
  char nhost[(HOSTLEN + 1)];

  /* Sanity check */
  if (!host)
    return -1;

  /* Copy host to a temp. array */
  ircd_strncpy(nhost,host,HOSTLEN);

  /* If current char. in hostmask is not alphanumeric AND isn't "-" or ".", it's
   * invalid according to DNS RFC - so disallow
   */
#ifndef ALLOW_NONRFC_HOSTCHARS
  for (s = nhost; *s; s++)
    if (!isalpha(*s) && !isdigit(*s) && !(*s == '.') && !(*s == '-') && !(*s == '~'))
       return 0;
#endif

  return -1;
}


/* --------------------------------------------------------
 * **Daz** GetVhostPass()
 *
 * Takes a hostmask string as a parameter and returns the
 * password for the corresponding S:line, NULL if not found
 * or if no password exists (oper only S:line)
 * --------------------------------------------------------
 */
char *GetVhostPass(char *hostmsk)
{
   struct ConfItem *aconf;

   for (aconf = GlobalConfList; aconf; aconf = aconf->next)
     if (aconf->status == CONF_VHOST && strCasediff(hostmsk, aconf->host) == 0)
       if (aconf->passwd)       /* The password field */
          return aconf->passwd;
       else
          return NULL;

   return NULL;
}


/* --------------------------------------------------------
 * **Daz** RejoinChannels()
 *
 * This function takes as its parameters a client and a string
 * and JOIN's the client to the channels in 'string'.  It is
 * effectively only a wrapper for the m_join public function.
 *
 *    (standard C boolean result)
 *    Returns:  -1 = Success  (user joined)
 *               0 = Failure (user couldn't join)
 * --------------------------------------------------------
 */
signed int RejoinChannels(struct Client *sptr, char *chanlist)
{
   struct Channel *chptr;

   char chanStr[(strlen(chanlist) + 1)];
   char keyStr[((MAXCHANNELSPERUSER * KEYLEN) + 1)];

   char *chanptr[3];
   char *channel;
   char *p = NULL;
   signed int res = -1;
   int argp = 2;

   chanStr[0] = '\0';
   keyStr[0] = '\0';

   for (channel = ircd_strtok(&p, chanlist, ","); channel; channel = ircd_strtok(&p, NULL, ",")) {
        /* Note: we can't check to see whether the channel is actually valid by using FindChannel,
         * since it is possible that the channel would cease to exist after having left it - and
         * therefore FindChannel would return NULL falsely.. what we do instead is just check to
         * make sure each 'channel' is a valid channel name..
         */
        if (IsChannelName(channel)) {
           if (chptr = FindChannel(channel)) {
              if (*chptr->mode.key) {
                 /* Make sure we don't overflow the buffer, since the keyStr[] variable length is
                  * an unknown quantity (since Opers can technically join any number of channels we
                  * cannot arbitrarily allocate memory for this field)
                  */
                 if ((strlen(keyStr) + strlen(chptr->mode.key)) < (MAXCHANNELSPERUSER + KEYLEN)) {
                    strncat(keyStr, chptr->mode.key, strlen(chptr->mode.key));
                    strcat(keyStr,",");
                    argp = 3;
                 }
              }
              strncat(chanStr, chptr->chname, strlen(chptr->chname));
              strcat(chanStr,",");
           } else {
              /* New channel (no key obviously) */
              strncat(chanStr, channel, strlen(channel));
              strcat(chanStr,",");
           }

       }
    }
    if (EmptyString(chanStr))    /* No channels to rejoin? */
       return 0;

    /* Cut off the end commas */
    if (strlen(keyStr) > 0)
       keyStr[(strlen(keyStr) - 1)] = '\0';
    chanStr[(strlen(chanStr) - 1)] = '\0';

    chanptr[0] = sptr->name;     /* parv[0] to m_join (sender prefix) */
    chanptr[1] = chanStr;        /* parv[1] to m_join (channel) */
    chanptr[2] = keyStr;         /* parv[2] to m_join (key(s)) */

    res = m_join(&me, sptr, argp, chanptr);

    return res;
}


